package com.scottlogic.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Random;

import org.junit.Test;
import static org.junit.Assert.*;

public class SortedListTest {

	//Comparator for comparing integers..
	Comparator<Integer> comp = new Comparator<Integer>(){
		public int compare(Integer one, Integer two){
			return one.intValue() - two.intValue();
		}
	};

	//Tests adding some elements to the end of the list then removing
	//at the front middle and end..
	@Test
	public void addingToEndAndRemoving(){

		List<Integer> ul = new SortedList<Integer>(comp);
		List<Integer> al = new ArrayList<Integer>();
		for(int i = 0; i < 6; i++){
			al.add(i);
			ul.add(i);
		}

		Iterator<Integer> ulItr = ul.iterator();
		Iterator<Integer> alItr = al.iterator();

		alItr.next();
		ulItr.next();
		int a = alItr.next();
		System.out.println(a);
		alItr.remove();
		int u = ulItr.next();
		System.out.println(u);
		ulItr.remove();

		//remove at the end..
		while(ulItr.hasNext()){
			ulItr.next();
			alItr.next();
		}
		ulItr.remove();
		alItr.remove();

		//see if the lists are the same..
		while(alItr.hasNext()){
			assertEquals(alItr.next(), ulItr.next());
		}
	}

	//Tests that the removeAl method works...
	@Test
	public void removeAll(){
		SortedList<Integer> sl = new SortedList<Integer>(comp);
		ArrayList<Integer> toRetain = new ArrayList<Integer>();
		//first just it fill..
		for(int i = 0; i < 10; i++){
			sl.add(i);
			sl.add(i);
			toRetain.add(i);
		}

		Collection<Integer> toRemove = new ArrayList<Integer>();
		toRemove.add(0);
		toRemove.add(5);
		toRemove.add(9);
		toRetain.removeAll(toRemove);

		//perform the removal, check it does something, then check it doesn't next time..
		boolean altered = sl.removeAll(toRemove);
		assertTrue(altered);
		altered = sl.removeAll(toRemove);
		assertFalse(altered);

		for(int i : toRemove){
			assertFalse(sl.contains(i));
		}
		for(Integer i : toRetain){
			assertTrue(sl.contains(i));
			sl.remove(i);
			assertTrue(sl.contains(i)); //the second element should remain..
		}

		//now see if removeAll works when there aren't any elements in the list..
		sl = new SortedList<Integer>(comp);
		altered = sl.removeAll(toRemove);
		assertFalse(altered);
	}

	//Tests that the removeAl method works...
	@Test
	public void retainAll(){

		SortedList<Integer> sl = new SortedList<Integer>(comp);
		ArrayList<Integer> toRemove = new ArrayList<Integer>();
		//first just it fill..
		for(int i = 0; i < 10; i++){
			sl.add(i);
			sl.add(i);
			toRemove.add(i);
		}

		Collection<Integer> toRetain = new ArrayList<Integer>();
		toRetain.add(0);
		toRetain.add(5);
		toRetain.add(9);
		toRemove.removeAll(toRetain);

		//run the retainAll method, check it does something, then check it doesn't next time..
		boolean altered = sl.retainAll(toRetain);
		assertTrue(altered);
		altered = sl.retainAll(toRetain);
		assertFalse(altered);

		//check it contains what we expect..
		for(int i : toRetain){
			assertTrue(sl.contains(i));
		}
		for(int i : toRemove){
			assertFalse(sl.contains(i));
		}
		for(Integer i : toRetain){
			assertTrue(sl.contains(i));
			sl.remove(i);
			assertTrue(sl.contains(i)); //the second element should remain..
		}

		//now see if retainAll works when there aren't any elements in the list..
		sl = new SortedList<Integer>(comp);
		altered = sl.removeAll(toRemove);
		assertFalse(altered);
	}

    //removes the root of the tree so that it needs to be rebalanced..
    @Test
    public void removeRoot(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(30);
        list.add(15);
        list.add(40);
        list.add(10);
        list.add(20);
        list.add(35);
        list.add(3);
        list.remove(new Integer(30));

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(15);
        test.add(10);
        test.add(35);
        test.add(3);
        test.add(20);
        test.add(40);

        assertTrue(list.structurallyEqualTo(test));
    }

    //removes the root of the tree so that it needs to be rebalanced..
    @Test
    public void removeNonRootWithTwoChildren(){
        //first where no rebalancing is required..
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(30);
        list.add(15);
        list.add(40);
        list.add(10);
        list.add(20);
        list.add(35);
        list.add(73);
        list.add(3);
        list.add(36);
        list.add(60);
        list.remove(new Integer(40));

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(30);
        test.add(15);
        test.add(60);
        test.add(10);
        test.add(20);
        test.add(35);
        test.add(73);
        test.add(3);
        test.add(36);

        assertTrue(list.structurallyEqualTo(test));

        //second where rebalancing is required..
        list.clear();
        list.add(30);
        list.add(15);
        list.add(40);
        list.add(10);
        list.add(20);
        list.add(35);
        list.add(73);
        list.add(3);
        list.add(36);
        list.add(60);
        list.add(83);
        list.add(79);
        list.remove(new Integer(40));

        test.clear();
        test.add(30);
        test.add(15);
        test.add(60);
        test.add(10);
        test.add(20);
        test.add(35);
        test.add(79);
        test.add(3);
        test.add(36);
        test.add(73);
        test.add(83);

        assertTrue(list.structurallyEqualTo(test));
    }

    //removes a left leaf where no rebalancing is required..
    @Test
    public void removeLeaf(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(30);
        list.add(15);
        list.add(40);
        list.add(10);
        list.add(20);
        list.add(35);
        list.add(73);
        list.add(3);
        list.add(36);
        list.add(60);
        list.remove(new Integer(60));

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(30);
        test.add(15);
        test.add(40);
        test.add(10);
        test.add(20);
        test.add(35);
        test.add(73);
        test.add(3);
        test.add(36);

        assertTrue(list.structurallyEqualTo(test));
    }

    //removes a leaf node where no rebalancing is required..
    @Test
    public void removeLeafWithRebalance(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(30);
        list.add(15);
        list.add(40);
        list.add(10);
        list.add(20);
        list.add(35);
        list.add(73);
        list.add(3);
        list.add(36);
        list.add(60);
        list.remove(new Integer(20));

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(30);
        test.add(10);
        test.add(40);
        test.add(3);
        test.add(15);
        test.add(35);
        test.add(73);
        test.add(36);
        test.add(60);

        assertTrue(list.structurallyEqualTo(test));
    }

    //removes a node that has a left child but no right one and no rebalancing is required..
    @Test
    public void removeWithJustLeftChild(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(30);
        list.add(15);
        list.add(40);
        list.add(10);
        list.add(20);
        list.add(35);
        list.add(73);
        list.add(3);
        list.add(36);
        list.add(60);
        list.remove(new Integer(73));

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(30);
        test.add(15);
        test.add(40);
        test.add(10);
        test.add(20);
        test.add(35);
        test.add(60);
        test.add(3);
        test.add(36);

        assertTrue(list.structurallyEqualTo(test));
    }

    //removes a node that has a right child but no left one and no rebalancing is required..
    @Test
    public void removeWithJustRightChild(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(30);
        list.add(15);
        list.add(40);
        list.add(10);
        list.add(20);
        list.add(35);
        list.add(73);
        list.add(3);
        list.add(36);
        list.add(85);
        list.remove(new Integer(73));

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(30);
        test.add(15);
        test.add(40);
        test.add(10);
        test.add(20);
        test.add(35);
        test.add(85);
        test.add(3);
        test.add(36);

        assertTrue(list.structurallyEqualTo(test));
    }

    //builds a tree where one left rotation is required..
    @Test
    public void addWhereLeftRotateRequired(){
        //first test when change happens at the root..
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(1);
        list.add(2);
        list.add(3);

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));

        //second test where not at the root..
        list.clear();
        list.add(0);
        list.add(-1);
        list.add(1);
        list.add(2);
        list.add(3);

        test.clear();
        test.add(0);
        test.add(-1);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));
    }

    //builds a tree where a double left rotation is required..
    @Test
    public void addWhereDoubleLeftRotateRequired(){
        //first test where change happens as root..
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(1);
        list.add(3);
        list.add(2);

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));

        //second test where problem isn't at the root..
        list.clear();
        list.add(0);
        list.add(-1);
        list.add(1);
        list.add(3);
        list.add(2);

        test.clear();
        test.add(0);
        test.add(-1);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));
    }

    //builds a tree where one right rotation is required..
    @Test
    public void addWhereRightRotateRequired(){
        //first test where needs to be done at root..
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(3);
        list.add(2);
        list.add(1);

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));

        //second test where not at the root..
        list.clear();
        list.add(0);
        list.add(-1);
        list.add(3);
        list.add(2);
        list.add(1);

        test.clear();
        test.add(0);
        test.add(-1);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));
    }

    //builds a tree where a double right rotation is required..
    @Test
    public void addWhereDoubleRightRotateRequired(){
        //first test where change is at root..
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(3);
        list.add(1);
        list.add(2);

        SortedList<Integer> test = new SortedList<Integer>(comp);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));

        //second test where it's not at the root..
        list.clear();
        list.add(4);
        list.add(5);
        list.add(3);
        list.add(1);
        list.add(2);

        test.clear();
        test.add(4);
        test.add(5);
        test.add(2);
        test.add(1);
        test.add(3);

        assertTrue(list.structurallyEqualTo(test));
    }

    //Sees if the contains method works as expected..
    @Test
    public void contains(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(1);
        list.add(2);
        list.add(3);
        list.add(4);
        list.add(5);
        list.add(6);
        list.add(7);
        list.add(8);
        list.add(7);
        list.add(8);

        for(int i = 1; i <= 8; i++){
            assertTrue(list.contains(i));
        }
        assertTrue(!list.contains(9));
    }

    //Tests the clear method..
    @Test
    public void clear(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(1);
        list.add(2);
        list.add(3);
        list.clear();

        assertTrue(!list.contains(1));
    }

    //Tests the get methods..
    @Test
    public void get(){
        SortedList<Integer> list = new SortedList<Integer>(comp);

        for(int i = 0; i < 100; i++){
            for(int j = 0; j < 3; j++){
                list.add(i);
            }
        }

        for(int i = 0; i < 300; i++){
            assertEquals(i/3, (long) list.get(i));
        }
    }

    //Tests the size methods..
    @Test
    public void size_with_lots_of_elements(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 5; i++){
            for(int j = 0; j < 1; j++){
                list.add(i);
            }
        }
        assertEquals(5, (long) list.size());
    }

    @Test
    public void size_with_1_elem(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(1);

        assertEquals(1, list.size());
    }

    @Test
    public void size_on_empty_list(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        assertEquals(0, list.size());
    }

    @Test
    public void size_on_list_with_2_elems(){
        //with new child on right..
    	SortedList<Integer> list = new SortedList<Integer>(comp);
        list.add(1);
        list.add(2);

        assertEquals(2, list.size());

        //with new child on left..
        list = new SortedList<Integer>(comp);
        list.add(2);
        list.add(1);

        assertEquals(2, list.size());
    }

    //Tests the toArray() method..
    @Test
    public void toObjectArray(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 100; i++){
            for(int j = 0; j < 3; j++){
                list.add(i);
            }
        }

        System.out.println("Size: " + list.size());

        Object [] array = list.toArray();
        assertTrue(array.length == 300);

        for(int i = 0; i < 300; i++){
            assertEquals((long) ((Integer) array[i]), (long) i/3);
        }
    }

    //Tests the toArray(T[] array) method..
    @Test
    public void toArrayWithType(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 100; i++){
            for(int j = 0; j < 3; j++){
                list.add(i);
            }
        }

        //test where array is too small..
        Integer [] array = list.toArray(new Integer[0]);
        assertTrue(array.length == 300);

        for(int i = 0; i < 300; i++){
            assertEquals((long) ((Integer) array[i]), (long) i/3);
        }

        //test where array is same size..
        array = list.toArray(new Integer[300]);
        assertTrue(array.length == 300);

        for(int i = 0; i < 300; i++){
            assertEquals((long) ((Integer) array[i]), (long) i/3);
        }

        //test where array is too big..
        array = list.toArray(new Integer[1000]);
        array[300] = -1;
        assertTrue(array.length == 1000);

        for(int i = 0; i < 300; i++){
            assertEquals((long) ((Integer) array[i]), (long) i/3);
        }
        assertEquals((long) array[300], -1);
    }

    //Tests the isEmpty method..
    //@Test
    public void isEmpty(){
        assertTrue(new SortedList<Integer>(comp ).isEmpty());
    }

    //Tests removing elements using their index..
    @Test
    public void removeWithIndex(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 100; i++){
            list.add(i);
        }
        for(int i = 99; i >= 0; i--){
            list.remove(i);
        }
        assertTrue(list.isEmpty());
    }

    //Tests removing values using their value..
    @Test
    public void removeWithValue(){
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 100; i++){
            list.add(i);
        }
        for(int i = 0; i < 100; i++){
            list.remove(new Integer(i));
        }
        assertTrue(list.isEmpty());
    }

    //Tests whether the tree remains balanced after many additions and removals.
    @Test
    public void checkBalance() {
        //build a random tree..
        Random rand = new Random(0);
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 10000; i++){
            list.add(rand.nextInt(1000));
        }
        for(int i = 0; i < 400; i++){
            list.remove(i);
        }

        System.out.println("MinBalance: " + list.minBalanceFactor());
        System.out.println("MaxBalance: " + list.maxBalanceFactor());

        assertTrue(list.minBalanceFactor() > -2);
        assertTrue(list.maxBalanceFactor() < 2);
    }

    @Test
    public void iterator_through_the_whole_list(){
    	SortedList<Integer> list = new SortedList<Integer>(comp);
    	for(int i = 0; i < 10000; i++){
            list.add(i);
        }
    	Iterator<Integer> itr = list.iterator();
    	int i = 0;
    	while(itr.hasNext()){
    		assertEquals(new Integer(i), itr.next());
    		i++;
    	}
    }

    @Test
    public void iterator_over_empty_list(){
    	SortedList<Integer> list = new SortedList<Integer>(comp);
    	Iterator<Integer> itr = list.iterator();
    	while(itr.hasNext()){
    		itr.next();
    	}
    }

    @Test
    public void iterator_remove_all(){
    	SortedList<Integer> list = new SortedList<Integer>(comp);
    	for(int j = 0; j < 3; j++){
	    	for(int i = 0; i < 6; i++){
		    	list.add(i);
	    	}
    	}

    	Iterator<Integer> itr = list.iterator();
    	itr.next();
    	itr.remove();
    	assertEquals(new Integer(0), itr.next());

    	itr.remove();
    	while(itr.hasNext()){
    		itr.next();
    		itr.remove();
		}
    	assertTrue(list.isEmpty());
    }

    //Tests if the itr remove method works as expected..
    @SuppressWarnings({"unchecked", "rawtypes"})
    @Test
    public void iteratorRemove(){
        //build a random tree..
        Random rand = new Random(1002);
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 100000; i++){
            list.add(rand.nextInt(10));
        }
        //Get copy of it..
        List copy = new ArrayList(Arrays.asList(list.toArray()));

        //remove 1st, 3rd and last elements..
        Iterator itr = list.iterator();
        itr.next();
        itr.remove();
        itr.next();
        itr.next();
        itr.remove();
        while(itr.hasNext()){
            itr.next();
        }
        itr.remove();

        //do the same with the copy..
        copy.remove(0);
        copy.remove(1);
        copy.remove(copy.size()-1);

        assertEquals(copy.size(), list.size());

        for(int i = 0; i < copy.size(); i++){
            assertEquals(copy.get(i), list.get(i));
        }
    }

    @Test
    public void listIteraterPrevious(){
    	SortedList<Integer> list = new SortedList<Integer>(comp);
    	for(int i = 0; i < 100; i++){
            list.add(i);
        }
    	ListIterator<Integer> listItr = list.listIterator(100);
    	int i = 0;
    	while(listItr.hasNext()){
    		assertEquals(new Integer(99-i), listItr.hasPrevious());
    	}
    }

    @Test (expected=UnsupportedOperationException.class)
    public void listIteraterSet(){
    	SortedList<Integer> list = new SortedList<Integer>(comp);
    	for(int i = 0; i < 100; i++){
            list.add(i);
        }
    	ListIterator<Integer> itr = list.listIterator(0);
    	itr.next();
    	itr.set(5);
    }

    @Test (expected=UnsupportedOperationException.class)
    public void listIteraterAdd(){
    	SortedList<Integer> list = new SortedList<Integer>(comp);
    	for(int i = 0; i < 100; i++){
            list.add(i);
        }
    	ListIterator<Integer> itr = list.listIterator(0);
    	itr.next();
    	itr.add(5);
    }

    //Tests if the itr throws concurrent modification errors as expected..
    @SuppressWarnings({"rawtypes"})
    @Test
    public void iteratorModException() {
        //build a tree..
        SortedList<Integer> list = new SortedList<Integer>(comp);
        for(int i = 0; i < 100; i++){
            list.add(i);
        }

        //get an iterator for it, then modify it..
        Iterator itr = list.iterator();
        list.add(101);

        ConcurrentModificationException modException = null;
        try {
            itr.next();
        } catch(ConcurrentModificationException e){
            modException = e;
        }

        assertTrue(modException != null);
    }

    private static class IntWrapper implements Comparable<IntWrapper> {

    	final int compValue; //the value the comparator will work on..
    	final int otherValue; //another value.

    	IntWrapper(int compValue, int otherValue){
    		this.compValue = compValue;
    		this.otherValue = otherValue;
    	}

		@Override
		public int compareTo(IntWrapper other) {
			return compValue-other.compValue;
		}

		@Override
		public boolean equals(Object other){
			return ((IntWrapper) other).compValue == compValue
						&& ((IntWrapper) other).otherValue == otherValue;
		}
    }

    public void testEqualElements(){
    	SortedList<IntWrapper> list = new SortedList<IntWrapper>(new Comparator<IntWrapper>(){
			@Override
			public int compare(IntWrapper a, IntWrapper b) {
				return a.compareTo(b);
			}
    	});
    	for(int i = 0; i < 100; i++){
    		list.add(new IntWrapper(0,i));
    	}
    	//should use same order as added..
    	for(int i = 0; i < 100; i++){
    		list.get(i).equals(new IntWrapper(0,i));
    	}
    }
}
